# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['urlstd']

package_data = \
{'': ['*']}

install_requires = \
['icupy>=0.11.0']

setup_kwargs = {
    'name': 'urlstd',
    'version': '2021.10.25.dev1',
    'description': 'Python implementation of the WHATWG URL Standard',
    'long_description': '# url-standard\n\n`urlstd` is a Python implementation of the WHATWG [URL Standard](https://url.spec.whatwg.org/).\n\nThis library provides `URL` class, `URLSearchParams` class, and low-level APIs that comply with the URL specification.\n\n## API\n\n- [URL class](https://url.spec.whatwg.org/#url-class)\n\n  - class urlstd.parse.`URL(url: str, base: Optional[str] = None)`\n    - [href](https://url.spec.whatwg.org/#dom-url-href): `readonly property href: str`\n    - [origin](https://url.spec.whatwg.org/#dom-url-origin): `readonly property origin: str`\n    - [protocol](https://url.spec.whatwg.org/#dom-url-protocol): `property protocol: str`\n    - [username](https://url.spec.whatwg.org/#dom-url-username): `property username: str`\n    - [password](https://url.spec.whatwg.org/#dom-url-password): `property password: str`\n    - [host](https://url.spec.whatwg.org/#dom-url-host): `property host: str`\n    - [hostname](https://url.spec.whatwg.org/#dom-url-hostname): `property hostname: str`\n    - [port](https://url.spec.whatwg.org/#dom-url-port): `property port: str`\n    - [pathname](https://url.spec.whatwg.org/#dom-url-pathname): `property pathname: str`\n    - [search](https://url.spec.whatwg.org/#dom-url-search): `property search: str`\n    - [searchParams](https://url.spec.whatwg.org/#dom-url-searchparams): `readonly property search_params: URLSearchParams`\n    - [hash](https://url.spec.whatwg.org/#dom-url-hash): `property hash: str`\n\n- [URLSearchParams class](https://url.spec.whatwg.org/#interface-urlsearchparams)\n\n  - class urlstd.parse.`URLSearchParams(init: Optional[Union[str, Sequence[Sequence[Union[str, int, float]]], Dict[str, Union[str, int, float]], URLRecord, URLSearchParams]] = None)`\n    - [append](https://url.spec.whatwg.org/#dom-urlsearchparams-append): `append(name: str, value: Union[str, int, float]) -> None`\n    - [delete](https://url.spec.whatwg.org/#dom-urlsearchparams-delete): `delete(name: str) -> None`\n    - [get](https://url.spec.whatwg.org/#dom-urlsearchparams-get): `get(name: str) -> Optional[str]`\n    - [getAll](https://url.spec.whatwg.org/#dom-urlsearchparams-getall): `get_all(name: str) -> Tuple[str, ...]`\n    - [has](https://url.spec.whatwg.org/#dom-urlsearchparams-has): `has(name: str) -> bool`\n    - [set](https://url.spec.whatwg.org/#dom-urlsearchparams-set): `set(name: str, value: Union[str, int, float]) -> None`\n    - [sort](https://url.spec.whatwg.org/#dom-urlsearchparams-sort): `sort() -> None`\n\n- Low-level APIs\n\n  - [URL parser](https://url.spec.whatwg.org/#concept-url-parser)\n\n    - urlstd.parse.`parse_url(urlstring: str, base: str = None, encoding: str = "utf-8") -> URLRecord`\n\n  - [basic URL parser](https://url.spec.whatwg.org/#concept-basic-url-parser)\n\n    - class urlstd.parse.`BasicURLParser`\n      - classmethod `parse(urlstring: str, base: Optional[URLRecord] = None, encoding: str = "utf-8", url: Optional[URLRecord] = None, state_override: Optional[URLParserState] = None) -> URLRecord`\n\n  - [URL record](https://url.spec.whatwg.org/#concept-url)\n\n    - class urlstd.parse.`URLRecord`\n      - [scheme](https://url.spec.whatwg.org/#concept-url-scheme): `property scheme: str = ""`\n      - [username](https://url.spec.whatwg.org/#concept-url-username): `property username: str = ""`\n      - [password](https://url.spec.whatwg.org/#concept-url-password): `property password: str = ""`\n      - [host](https://url.spec.whatwg.org/#concept-url-host): `property host: Optional[Union[str, int, Tuple[int, ...]]] = None`\n      - [port](https://url.spec.whatwg.org/#concept-url-port): `property port: Optional[int] = None`\n      - [path](https://url.spec.whatwg.org/#concept-url-path): `property path: Union[List[str], str] = []`\n      - [query](https://url.spec.whatwg.org/#concept-url-query): `property query: Optional[str] = None`\n      - [fragment](https://url.spec.whatwg.org/#concept-url-fragment): `property fragment: Optional[str] = None`\n      - [origin](https://url.spec.whatwg.org/#concept-url-origin): `readonly property origin: Optional[Origin]`\n      - [is special](https://url.spec.whatwg.org/#is-special): `is_special() -> bool`\n      - [is not special](https://url.spec.whatwg.org/#is-not-special): `is_not_special() -> bool`\n      - [includes credentials](https://url.spec.whatwg.org/#include-credentials): `includes_credentials() -> bool`\n      - [has an opaque path](https://url.spec.whatwg.org/#url-opaque-path): `has_opaque_path() -> bool`\n      - [cannot have a username/password/port](https://url.spec.whatwg.org/#cannot-have-a-username-password-port): `cannot_have_username_password_port() -> bool`\n      - [URL serializer](https://url.spec.whatwg.org/#concept-url-serializer): `serialize_url(exclude_fragment: bool = False) -> str`\n      - [host serializer](https://url.spec.whatwg.org/#concept-host-serializer): `serialize_host() -> str`\n      - [URL path serializer](https://url.spec.whatwg.org/#url-path-serializer): `serialize_path() -> str`\n\n  - IDNA\n\n    - class urlstd.parse.`IDNA`\n      - [domain to ASCII](https://url.spec.whatwg.org/#concept-domain-to-ascii): classmethod `domain_to_ascii(domain: str, be_strict: bool = False) -> str`\n\n  - Host parsing/Host serializing\n\n    - class urlstd.parse.`Host`\n      - [host parser](https://url.spec.whatwg.org/#concept-host-parser): classmethod `parse(host: str, is_not_special: bool = False) -> Union[str, int, Tuple[int, ...]]`\n      - [host serializer](https://url.spec.whatwg.org/#concept-host-serializer): classmethod `serialize(host: Union[str, int, Sequence[int]]) -> str`\n\n  - [percent-decode a string](https://url.spec.whatwg.org/#string-percent-decode)\n\n    - urlstd.parse.`string_percent_decode(s: str) -> bytes`\n\n  - [percent-encode after encoding](https://url.spec.whatwg.org/#string-percent-encode-after-encoding)\n\n    - urlstd.parse.`string_percent_encode(s: str, safe: str, encoding: str = "utf-8", space_as_plus: bool = False) -> str`\n\n  - [application/x-www-form-urlencoded parser](https://url.spec.whatwg.org/#concept-urlencoded-parser)\n\n    - urlstd.parse.`parse_qsl(query: bytes) -> List[Tuple[str, str]]`\n\n  - [application/x-www-form-urlencoded serializer](https://url.spec.whatwg.org/#concept-urlencoded-serializer)\n\n    - urlstd.parse.`urlencode(query: Sequence[Tuple[str, str]], encoding: str = "utf-8") -> str`\n\n- Compatibility with standard library `urllib`\n\n  - urlstd.parse.`urlparse(urlstring: str, base: str = None, encoding: str = "utf-8", allow_fragments: bool = True) -> urllib.parse.ParseResult`\n\n    An alternative to `urllib.parse.urlparse()`.\n    Parses a string representation of a URL using the basic URL parser, and returns `urllib.parse.ParseResult`.\n\n## Basic Usage\n\n```python\n>>> from urlstd.parse import URL\n>>> url = URL(\'?ﬃ&🌈\', \'http://example.org\')\n>>> url\nURL(href=\'http://example.org/?%EF%AC%83&%F0%9F%8C%88\', origin=\'http://example.org\', protocol=\'http:\', username=\'\', password=\'\', host=\'example.org\', hostname=\'example.org\', port=\'\', pathname=\'/\', search=\'?%EF%AC%83&%F0%9F%8C%88\', hash=\'\')\n>>> url.search\n\'?%EF%AC%83&%F0%9F%8C%88\'\n>>> params = url.search_params\n>>> params\nURLSearchParams([(\'ﬃ\', \'\'), (\'🌈\', \'\')])\n>>> params.sort()\n>>> params\nURLSearchParams([(\'🌈\', \'\'), (\'ﬃ\', \'\')])\n>>> url.search\n\'?%F0%9F%8C%88=&%EF%AC%83=\'\n>>> str(url)\n\'http://example.org/?%F0%9F%8C%88=&%EF%AC%83=\'\n```\n\nAn alternative to urllib.parse.urlparse():\n\n```python\n>>> import html\n>>> from urllib.parse import unquote\n>>> from urlstd.parse import urlparse\n>>> urlparse(\'?aÿb\', \'http://example.org/foo/\', encoding=\'utf-8\')\nParseResult(scheme=\'http\', netloc=\'example.org\', path=\'/foo/\', params=\'\', query=\'a%C3%BFb\', fragment=\'\')\n>>> unquote(\'a%C3%BFb\')\n\'aÿb\'\n>>> urlparse(\'?aÿb\', \'http://example.org/foo/\', encoding=\'windows-1251\')\nParseResult(scheme=\'http\', netloc=\'example.org\', path=\'/foo/\', params=\'\', query=\'a%26%23255%3Bb\', fragment=\'\')\n>>> unquote(\'a%26%23255%3Bb\', encoding=\'windows-1251\')\n\'a&#255;b\'\n>>> html.unescape(\'a&#255;b\')\n\'aÿb\'\n>>> urlparse(\'?aÿb\', \'http://example.org/foo/\', encoding=\'windows-1252\')\nParseResult(scheme=\'http\', netloc=\'example.org\', path=\'/foo/\', params=\'\', query=\'a%FFb\', fragment=\'\')\n>>> unquote(\'a%FFb\', encoding=\'windows-1252\')\n\'aÿb\'\n```\n\n## Logging\n\n`urlstd` uses standard library `logging` for [validation error](https://url.spec.whatwg.org/#validation-error).\nChange the logger log level of urlstd if needed:\n\n```python\nlogging.getLogger("urlstd").setLevel(logging.ERROR)\n```\n\n## Dependencies\n\n- [icupy](https://pypi.org/project/icupy/) >= 0.11.0\n  - [ICU](https://icu.unicode.org/download) - International Components for Unicode (ICU4C)\n\n## Installation\n\n1. Configuring environment variables for icupy (ICU):\n\n    - Windows:\n\n      - Set the `ICU_ROOT` environment variable to the root of the ICU installation (default is `C:\\icu`). For example, if the ICU is located in `C:\\icu4c`:\n\n        ```bat\n        set ICU_ROOT=C:\\icu4c\n        ```\n\n        or in PowerShell:\n\n        ```bat\n        $env:ICU_ROOT = "C:\\icu4c"\n        ```\n\n    - Linux/POSIX:\n\n      - If the ICU is located in a non-regular place, set the `PKG_CONFIG_PATH` and `LD_LIBRARY_PATH` environment variables. For example, if the ICU is located in `/usr/local`:\n\n        ```bash\n        export PKG_CONFIG_PATH=/usr/local/lib/pkgconfig:$PKG_CONFIG_PATH\n        export LD_LIBRARY_PATH=/usr/local/lib:$LD_LIBRARY_PATH\n        ```\n\n2. Installing from PyPI:\n\n    ```bash\n    pip install urlstd\n    ```\n\n## Running Tests\n\nTo run tests and generate a report:\n\n```bash\ngit clone https://github.com/miute/url-standard.git\ncd url-standard\ntox -e wpt\n```\n\nSee result: [tests/wpt/report.html](https://htmlpreview.github.io/?https://github.com/miute/url-standard/blob/main/tests/wpt/report.html)\n\n## License\n\n[MIT License](https://github.com/miute/url-standard/blob/main/LICENSE).\n',
    'author': 'Tetsuya Miura',
    'author_email': 'miute.dev@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/miute/url-standard',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
